"""GE calculation for domains and chains."""
import ast
import csv
import json
import pathlib
import sys
import time
import warnings

import numpy as np
import pyge.gent as gent
from Bio import SeqIO
from Bio.PDB.PDBExceptions import PDBConstructionWarning
from pyge.contacts.pdb_parser import get_residues
from res2ca import res2ca
from save_ge import save_config, save_loops

sys.path.append("/home/ubuntu/membrane/")
from stats_dataset.utils import domain_segments_parser


if len(sys.argv) != 4:
    print("Usage: python ge.py <starting_idx> <ending_idx/'end'> <id>")
    sys.exit(1)
elif len(sys.argv) > 1:
    batch_start = int(sys.argv[1])
    if sys.argv[2] == "end":
        batch_end = "end"
    else:
        batch_end = int(sys.argv[2])
    out_id = sys.argv[3]

membrane_dir = pathlib.Path("/mnt/MEMBRANE/")
ali_dir = membrane_dir / "ALIGNMENTS"
cm_dir = membrane_dir / "CM"
ge_dir = membrane_dir / "GE"

############################################
# --- parameters ---
# Consider interdomain contacts, hence with the whole
# chain
ge4interdomain = False
# Cut domains with missing residues segments larger
# then a threshold
remove_modeled_segments_thr = True
domain2remove_path = membrane_dir / "CHAINS_FIX/LOG/chains_w_modeled_above_thr_10.csv"
############################################

# --- error log ---
err_file = open(ge_dir / "LOG" / f"err_log_{out_id}.csv", "w")
err_writer = csv.writer(err_file)
err_writer.writerow(["# generated by ge.py"])
err_writer.writerow(["# chain_id", "domain_id", "err_msg"])

# --- load list of chains (from CHAINS_FIX) ---
with open(ge_dir / "chains_list.txt", "r") as infile:
    reader = csv.reader(infile)
    next(reader)
    next(reader)
    chain_path_modres = []
    for row in reader:
        if len(row) == 3:
            chain_path_modres.append(
                [
                    row[0],
                    pathlib.Path(row[1]),
                    row[2],
                ]
            )
        elif len(row) == 4:
            chain_path_modres.append(
                [
                    row[0],
                    pathlib.Path(row[1]),
                    row[2],
                    ast.literal_eval(row[3]),
                ]
            )
        else:
            raise RuntimeError("Error during parsing chains_list.txt")
assert len(chain_path_modres) == 1282 - 1
# -1 -> see readme.md

# --- load info about modeled regions ---
# load alignments
ali_files = [f for f in ali_dir.iterdir() if f.suffix == ".ali"]
# create a dictionary chain:sequence (only for modeled chains)
chain_seq = {}
for ali in ali_files:
    chain_seq[ali.stem] = SeqIO.to_dict(SeqIO.parse(ali, "pir"))[ali.stem].seq

# --- load CATH info (chain_info.csv) ---
with open(membrane_dir / "chains_list.csv", "r") as file_cath:
    reader = csv.reader(file_cath)
    chain_domains = {}
    for row in reader:
        chain_id = f"{row[0][:4]}_{row[2]}"
        chain_domains[chain_id] = domain_segments_parser(
            json.loads(row[4].replace("'", "\""))
        )
assert len(chain_domains.keys()) == 1282

# --- remove regions modeled ---
if remove_modeled_segments_thr:
    with open(domain2remove_path, "r") as dom2rem_file:
        reader = csv.reader(dom2rem_file)
        next(reader)
        next(reader)
        next(reader)
        dom2rem = {}
        for row in reader:
            chain_id = row[0]
            boundaries = list(ast.literal_eval(row[2]))
            # 0-indexing
            boundaries = [b - 1 for b in boundaries]
            if chain_id not in dom2rem:
                dom2rem[chain_id] = [boundaries]
            else:
                dom2rem[chain_id].append(boundaries)

# --- loop over chains ---
if batch_end == 'end':
    batch_end = len(chain_path_modres)
for row in chain_path_modres[batch_start:batch_end]:
    # unpack row
    chain_id = row[0]
    chain_path = row[1]
    # here we assume that some pdb could have a unknow residue
    # yes, it happens
    if len(row) == 3:
        modres = ["UNK"]
    elif len(row) == 4:
        modres = row[3] + ["UNK"]
    else:
        raise RuntimeError("Error in main loop")
    print(f"====== Name: {chain_id} ======")
    start = time.time()

    # --- load CA coordinates ---
    with warnings.catch_warnings():
        # avoid warning regarding END record in pdbs
        warnings.simplefilter('ignore', PDBConstructionWarning)
        res_list = get_residues(
            str(chain_path),
            1,
            chain_id[-1:],
            to_include=modres,
            to_ignore=["HETATM"]
        )
    ca_coords = res2ca(res_list)
    # --- load CM ---
    try:
        with np.load(cm_dir / f"{chain_id}.npz") as data:
            chain_cm = data.f.arr
    except FileNotFoundError:
        print("CM not yet calculated")
        err_writer.writerow([chain_id, "", "CM not yet calculated"])
        continue
    # --- load Sequence info, identify tails ---
    if chain_id in chain_seq:
        sequence = chain_seq[chain_id]
        # INDEX of the first element in sequence to be different from '-'
        first_modeled_res_idx = next(i for i, x in enumerate(sequence) if x != "-")
        # INDEX of the last element in sequence to be different from '-'
        last_modeled_res_idx = len(sequence) - next(
            i for i, x in enumerate(reversed(sequence)) if x != "-"
        ) - 1  # -1 because of the 0-based indexing
    else:
        # In this case, a .ali file has not been produced. Therefore,
        # the chain has not been modeled
        first_modeled_res_idx = 0
        last_modeled_res_idx = chain_cm.shape[0] - 1
    
    print(f"first_modeled_res_idx: {first_modeled_res_idx}")
    print(f"last_modeled_res_idx: {last_modeled_res_idx}")

    # --- extract modeled segments with len > 10 ---
    if remove_modeled_segments_thr:
        if chain_id in dom2rem:
            boundaries_modeled_above_10 = dom2rem[chain_id]
        else:
            boundaries_modeled_above_10 = []

    # prepare cm for the rest of the chain
    inter_domain_loops_cm = chain_cm.copy()
    
    # --- loop over domains ---
    for domain_code, domain_boundaries in chain_domains[chain_id].items():
        print(f"+++ Domain: {domain_code} +++")
        # convert domain boundaries to 0-based indexing
        domain_boundaries = [
            [b[0] - 1, b[1] - 1] for b in domain_boundaries
        ]
        # check that boundaries do not contains modeled tails
        if domain_boundaries[0][0] < first_modeled_res_idx:
            print("WARNING: lower domain boundary is outside of modeled region")
            domain_boundaries[0][0] = first_modeled_res_idx
        if domain_boundaries[-1][1] > last_modeled_res_idx:
            print("WARNING: upper domain boundary is outside of modeled region")
            domain_boundaries[-1][1] = last_modeled_res_idx
        # initialize domain CM
        if len(domain_boundaries) == 1:
            dim_domain = domain_boundaries[0][1] - domain_boundaries[0][0] + 1
            print(f"Single segment domain; len = {dim_domain}")
        else:
            dim_domain = sum(
                (b[1] - b[0] + 1 for b in domain_boundaries)
            )
            print(f"Multiple segment domain; len = {dim_domain}")

        # --- check if domain has modeled segments > 10 ---
        if remove_modeled_segments_thr:
            skip = False
            for b in boundaries_modeled_above_10:
                stop_inner = False
                for db in domain_boundaries:
                    if db[0] < b[0] and db[1] > b[1]:
                        print("Domain contains modeled segments w/ len > 10")
                        print(f"Domain boundaries: {domain_boundaries}; modeled boundaries: {b}")
                        skip = True
                        stop_inner = True
                        break
                if stop_inner:
                    break
            if skip:
                continue

        domain_cm = np.zeros((dim_domain, dim_domain))
        domain_ca_coords = np.zeros((dim_domain, 3))
        # --- select CA coords and CM blocks relative to the domain ---
        if len(domain_boundaries) == 1:
            domain_ca_coords = ca_coords[
                domain_boundaries[0][0]:domain_boundaries[0][1] + 1
            ]
            domain_cm = chain_cm[
                domain_boundaries[0][0]:domain_boundaries[0][1] + 1,
                domain_boundaries[0][0]:domain_boundaries[0][1] + 1,
            ]
            # remove intra-domain loops
            inter_domain_loops_cm[
                domain_boundaries[0][0]:domain_boundaries[0][1] + 1,
                domain_boundaries[0][0]:domain_boundaries[0][1] + 1,
            ] = 0
        else:
            domain_len_progress = 0
            for i, segment in enumerate(domain_boundaries):
                segment_len = segment[1] - segment[0] + 1
                domain_ca_coords[
                    domain_len_progress: domain_len_progress + segment_len
                ] = ca_coords[segment[0]: segment[1] + 1]
                domain_cm[
                    domain_len_progress: domain_len_progress + segment_len,
                    domain_len_progress: domain_len_progress + segment_len,
                ] = chain_cm[
                    segment[0]: segment[1] + 1, segment[0]: segment[1] + 1
                ]
                domain_len_progress += segment_len

                # remove intra-domain loops
                inter_domain_loops_cm[
                    segment[0]: segment[1] + 1, segment[0]: segment[1] + 1
                ] = 0
            if domain_len_progress != dim_domain:
                print("Domain length mismatch")
                err_writer.writerow([
                    chain_id,
                    "",
                    f"Domain length mismatch; domain_len_progress={domain_len_progress} dim_domain={dim_domain}"
                ])
                continue

        # --- domain GE ---
        try:
            ge_loops_domains = gent.ge_loops(
                domain_cm, domain_ca_coords, 10
            )
            save_loops(
                ge_dir / "DOMAINS" / "LOOPS" / f"{domain_code.upper()}.csv",
                ge_loops_domains
            )
            ge_domains_max = gent.ge_configuration(
                ge_loops_domains, 0, mode="max"
            )
            ge_domains_w = gent.ge_configuration(
                ge_loops_domains, 0, mode="weighted"
            )
            save_config(
                ge_dir / "DOMAINS" / "CONFIG" / f"{domain_code.upper()}.csv",
                ge_domains_max,
                ge_domains_w
            )
        except Exception as err:
            print(f"ERROR: {err}")
            err_writer.writerow([chain_id, domain_code, err])
            continue
    
    # --- chain GE ---
    if ge4interdomain:
        print("+++ Chain GE +++")
        # removing tails
        inter_domain_loops_cm_no_tails = inter_domain_loops_cm[
            first_modeled_res_idx:last_modeled_res_idx + 1,
            first_modeled_res_idx:last_modeled_res_idx + 1,
        ]
        ca_coords_no_tails = ca_coords[first_modeled_res_idx:last_modeled_res_idx + 1]
        # computation
        print(f"Chain len w/out tails: {inter_domain_loops_cm_no_tails.shape[0]}")
        if (inter_domain_loops_cm_no_tails != 0).sum() == 0:
            print("No loops remaining in chain")
        try:
            ge_loops_chain = gent.ge_loops(
                inter_domain_loops_cm_no_tails,
                ca_coords_no_tails, 10
            )
            save_loops(
                ge_dir / "CHAINS" / "LOOPS" / f"{chain_id}.csv",
                ge_loops_chain
            )
            ge_chain_max = gent.ge_configuration(
                ge_loops_chain, 0, mode="max"
            )
            ge_chain_w = gent.ge_configuration(
                ge_loops_chain, 0, mode="weighted"
            )
            save_config(
                ge_dir / "CHAINS" / "CONFIG" / f"{chain_id}.csv",
                ge_chain_max,
                ge_chain_w
            )
        except Exception as err:
            print(f"ERROR: {err}")
            err_writer.writerow([chain_id, "", err])
    print(f"\nTime (chain): {time.time() - start} s")
